
#include "pspscreen.h"

#include <stdlib.h>

#include <pspdisplay.h>
#include <pspgu.h>

static unsigned char *vram,*vramp,*active,*hidden;
static int bpp,bytes,fmt,gufmt,pages;
static unsigned int list[4096] __attribute__((aligned(32)));

void *screen_open(int bitsperpixel,int num_pages)
{
    switch(bitsperpixel)
    {
        case 15: bpp=15; bytes=2; fmt=PSP_DISPLAY_PIXEL_FORMAT_5551;
                                  gufmt=GU_PSM_5551; break;
        case 16: bpp=16; bytes=2; fmt=PSP_DISPLAY_PIXEL_FORMAT_565;
                                  gufmt=GU_PSM_5650; break;
        case 32: bpp=32; bytes=4; fmt=PSP_DISPLAY_PIXEL_FORMAT_8888;
                                  gufmt=GU_PSM_8888; break;
        default: return(NULL);
    }

    if(num_pages==0)
        return(NULL);
    pages=num_pages;

    vram=active=sceGeEdramGetAddr();
    vramp=vram+512*272*pages*bytes;
    hidden=vramp;
    vramp+=512*512*bytes;
    
    sceDisplaySetMode(0,480,272);
    sceDisplaySetFrameBuf(vram,512,fmt,PSP_DISPLAY_SETBUF_NEXTFRAME);

    sceGuInit();
    sceGuStart(GU_DIRECT,list);

    sceGuDrawBuffer(gufmt,vram,512);
    sceGuDispBuffer(480,272,vram,512);
    sceGuClear(GU_COLOR_BUFFER_BIT);
    sceGuViewport(2048,2048,480,272);
    sceGuOffset(2048-(2048/2),2048-(272/2));
    
    sceGuScissor(0,0,480,272);
    sceGuEnable(GU_SCISSOR_TEST);
    sceGuDisable(GU_DEPTH_TEST);
    sceGuFrontFace(GU_CW);
    sceGuFinish();
    sceGuSync(0, 0);
    sceGuDisplay(1);

    sceDisplaySetFrameBuf(vram,512,fmt,PSP_DISPLAY_SETBUF_NEXTFRAME);

    return(vram);
}

void screen_waitVB(void)
{
    sceDisplayWaitVblankStart();
}

void screen_vispage(int n)
{
    sceKernelDcacheWritebackAll();
    sceDisplaySetFrameBuf(vram+n*512*272*bytes,512,fmt,
                          PSP_DISPLAY_SETBUF_NEXTFRAME);
}

void *screen_drawpage(int n)
{
    active=vram+n*512*272*bytes;
    return(active);
}

void *screen_offscreen(void)
{
    return(hidden);
}

struct texVertex
{
	unsigned short u, v;
	short x, y, z;
};

void screen_blit(int x,int y,int width,int height,float aspect,void *buf)
{
    struct texVertex *v;

    sceKernelDcacheWritebackAll();

    if(buf!=hidden)
    {
        sceGuStart(GU_DIRECT,list);
        sceGuCopyImage(gufmt,0,0,width,height,512,buf,0,0,512,hidden);
        sceGuFinish();
        sceGuSync(0,0);
    }

    sceGuStart(GU_DIRECT,list);
    sceGuDrawBuffer(gufmt,active,512);
    sceGuDispBuffer(480,272,active,512);
    sceGuEnable(GU_TEXTURE_2D);
    sceGuTexMode(gufmt,0,0,0);
    sceGuTexFunc(GU_TFX_REPLACE, GU_TCC_RGB);
    sceGuTexFilter(GU_LINEAR, GU_LINEAR);
    sceGuTexImage(0,512,512,512,hidden);
    sceGuTexSync();

    v=(struct texVertex*)sceGuGetMemory(2*sizeof(struct texVertex));

    v[0].u=0;
    v[0].v=0;
    v[1].u=width-1;
    v[1].v=height-1;

    v[0].x=x;
    v[0].y=y;
    v[1].x=x+271*aspect;
    v[1].y=y+271;
    v[0].z=v[1].z=0;

    sceGuDrawArray(GU_SPRITES,GU_TEXTURE_16BIT|GU_VERTEX_16BIT|GU_TRANSFORM_2D,
			2,0,v);

    sceGuFinish();
    sceGuSync(0,0);
}
